const std = @import("std");

const sys = @import("sys.zig");

pub const FramebufferError = error{
    BppUnsupported,
};

pub const Framebuffer = struct {
    const Self = @This();

    file: std.fs.File,
    data: []align(std.mem.page_size) u8,

    xres: u32,
    yres: u32,
    bits_per_pixel: u32,

    /// Extracted from `eips` binary
    fn checkBppSupported(bpp: c_uint) bool {
        return (bpp <= 8) and
            ((@as(c_uint, 0x68800000) << @truncate(bpp)) >> ((@sizeOf(c_uint) * 8) - 1)) == 1;
    }

    pub fn open(path: []const u8) !Self {
        const file = try std.fs.cwd().openFile(path, .{
            .mode = .read_write,
        });

        const vinfo = try sys.vscreeninfo(file.handle);
        if (!checkBppSupported(vinfo.bits_per_pixel)) {
            return FramebufferError.BppUnsupported;
        }

        const data = try sys.mmap_fb(file.handle, vinfo);

        return .{
            .file = file,
            .data = data,
            .xres = vinfo.xres,
            .yres = vinfo.yres,
            .bits_per_pixel = vinfo.bits_per_pixel,
        };
    }

    pub fn deinit(self: *Self) void {
        std.posix.munmap(self.data);
        self.file.close();
    }

    pub fn fill(self: *Self, color: u8) void {
        @memset(self.data, color);
    }

    pub fn fillLine(self: *Self, y: u32, color: u8) void {
        if (y >= self.yres)
            return;

        const buf = self.data[self.offset(0, y)..self.offset(self.xres, y)];
        @memset(buf, color);
    }

    pub fn fillLineVar(self: *Self, y: u32, x_start: u32, x_end: u32, color: u8) void {
        if (y >= self.yres or
            x_start > x_end or
            x_start >= self.xres)
            return;

        if (x_start == x_end) {
            self.set(x_start, y, color);
            return;
        }

        const safe_x_end = @min(x_end, self.xres);
        const buf = self.data[self.offset(x_start, y)..self.offset(safe_x_end, y)];
        @memset(buf, color);
    }

    pub fn fillArea(self: *Self, region: sys.Region, color: u8) void {
        if (region.top >= self.yres or
            region.left >= self.xres or
            region.width == 0 or
            region.height == 0)
            return;

        const y_end = @min(region.top + region.height + 1, self.yres);
        const x_end = @min(region.left + region.width + 1, self.xres);

        for (region.top..y_end) |y| {
            const buf = self.data[self.offset(region.left, y)..self.offset(x_end, y)];
            @memset(buf, color);
        }
    }

    pub fn clear(self: *Self) !void {
        try sys.clear_screen(self.file.handle);
        self.fill(0x00);
    }

    pub fn update(self: *const Self, update_mode: sys.UpdateMode) !void {
        try sys.update_display(self.file.handle, update_mode);
    }

    pub fn updateArea(self: *const Self, region: sys.Region, update_mode: sys.UpdateMode) !void {
        if (region.width == self.xres and region.height == self.yres) {
            try sys.update_display(self.file.handle, update_mode);
        } else {
            try sys.update_display_area(self.file.handle, region, update_mode);
        }
    }

    inline fn offset(self: *const Self, x: usize, y: usize) usize {
        return y * self.xres + x;
    }

    pub fn get(self: *const Self, x: usize, y: usize) u8 {
        return self.data[self.offset(x, y)];
    }

    pub fn set(self: *Self, x: usize, y: usize, color: u8) void {
        if (x < self.xres and y < self.yres) {
            self.data[self.offset(x, y)] = color;
        }
    }
};
