const std = @import("std");

const c = @cImport({
    @cInclude("linux/fb.h");
    @cInclude("einkfb.h");
});

fn handleErrno(rc: usize) std.posix.UnexpectedError!void {
    const err = std.posix.errno(rc);
    if (err != .SUCCESS)
        return std.posix.unexpectedErrno(err);
}

pub fn vscreeninfo(fd: std.posix.fd_t) std.posix.UnexpectedError!c.fb_var_screeninfo {
    var vinfo: c.fb_var_screeninfo = undefined;
    try handleErrno(
        std.os.linux.ioctl(fd, c.FBIOGET_VSCREENINFO, @intFromPtr(&vinfo)),
    );

    return vinfo;
}

pub fn mmap_fb(fd: std.posix.fd_t, vinfo: c.fb_var_screeninfo) ![]align(std.mem.page_size) u8 {
    const data_size: usize = vinfo.xres * vinfo.yres * (vinfo.bits_per_pixel / 8);
    const data = try std.posix.mmap(
        null,
        data_size,
        std.posix.PROT.READ | std.posix.PROT.WRITE,
        .{
            .TYPE = .SHARED,
            .LOCKED = false,
        },
        fd,
        0,
    );

    return data;
}

pub fn clear_screen(fd: std.posix.fd_t) std.posix.UnexpectedError!void {
    try handleErrno(
        std.os.linux.ioctl(fd, c.FBIO_EINK_CLEAR_SCREEN, c.EINK_CLEAR_SCREEN),
    );
}

pub const UpdateMode = enum {
    full,
    partial,

    fn toFxType(self: UpdateMode) c.fx_type {
        switch (self) {
            .full => return c.fx_update_full,
            .partial => return c.fx_update_partial,
        }
    }
};

pub fn update_display(fd: std.posix.fd_t, update_mode: UpdateMode) std.posix.UnexpectedError!void {
    const fx = update_mode.toFxType();

    try handleErrno(
        std.os.linux.ioctl(fd, c.FBIO_EINK_UPDATE_DISPLAY, @intCast(fx)),
    );
}

pub fn Rect(comptime T: type) type {
    return struct {
        top: T,
        left: T,
        width: T,
        height: T,
    };
}

pub const Region = Rect(u32);

pub fn update_display_area(fd: std.posix.fd_t, region: Region, update_mode: UpdateMode) std.posix.UnexpectedError!void {
    const area: c.update_area_t = .{
        .x1 = @intCast(region.left),
        .y1 = @intCast(region.top),
        .x2 = @intCast(region.left + region.width),
        .y2 = @intCast(region.top + region.height),
        .which_fx = update_mode.toFxType(),
        .buffer = 0,
    };

    try handleErrno(
        std.os.linux.ioctl(fd, c.FBIO_EINK_UPDATE_DISPLAY_AREA, @intFromPtr(&area)),
    );
}
